'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.addImplicitReturns = addImplicitReturns;
exports.toPlainFunction = toPlainFunction;
exports.replaceWithPlainFunction = replaceWithPlainFunction;
exports.toArrowFunction = toArrowFunction;
exports.replaceWithArrowFunction = replaceWithArrowFunction;
exports.replaceWithBoundFunction = replaceWithBoundFunction;
exports.isNamedArrowFunction = isNamedArrowFunction;

var _is = require('./is');

var _is2 = _interopRequireDefault(_is);

var _tails = require('./tails');

var _types = require('./types');

var _types2 = _interopRequireDefault(_types);

var _blocks = require('./blocks');

var _getLoc = require('ast-loc-utils/lib/getLoc');

var _getLoc2 = _interopRequireDefault(_getLoc);

var _buildAtLoc = require('ast-loc-utils/lib/buildAtLoc');

var _buildAtLoc2 = _interopRequireDefault(_buildAtLoc);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

// Add "return" statements to all expressions in tail position inside the
// body of the function at the given Babel traversal path.
function addImplicitReturns(functionPath) {
  var _ref;

  var node = functionPath.node;
  // Don't add implicit returns to constructors, setters, or :void functions
  if (node.kind === "constructor" || node.kind === "set") return;
  if ((0, _is2.default)("VoidTypeAnnotation", (_ref = node.returnType) == null ? null : _ref.typeAnnotation)) return;

  return (0, _tails.transformTails)(functionPath.get("body"), false, false, function (expr) {
    return (0, _buildAtLoc2.default)((0, _getLoc2.default)(expr), _types2.default.returnStatement, expr);
  });
} // Convert any functional node (NamedArrow etc) back to a JS
// `FunctionDeclaration` or `FunctionExpression`


function toPlainFunction(node) {
  var id = node.id,
      params = node.params,
      body = node.body,
      generator = node.generator,
      async = node.async;


  var nextNode = (0, _is.isa)(node, "Statement") ? _types2.default.functionDeclaration(id, params, (0, _blocks.toBlockStatement)(body), generator, async) : _types2.default.functionExpression(id, params, (0, _blocks.toBlockStatement)(body), generator, async);

  if (node.returnType) nextNode.returnType = node.returnType;
  if (node.typeParameters) nextNode.typeParameters = node.typeParameters;

  return nextNode;
}

function replaceWithPlainFunction(path) {
  return path.replaceWith(toPlainFunction(path.node));
} // Convert any functional node (NamedArrow etc) back to a JS arrow function
// `ArrowFunctionExpression`


function toArrowFunction(node) {
  var id = node.id,
      params = node.params,
      body = node.body,
      async = node.async;


  if ((0, _is2.default)("Statement", node)) {
    var fn = _types2.default.arrowFunctionExpression(params, body, async);
    if (node.returnType) fn.returnType = node.returnType;
    if (node.typeParameters) fn.typeParameters = node.typeParameters;
    return _types2.default.variableDeclaration("const", [_types2.default.variableDeclarator(id, fn)]);
  } else {
    // just throw away the id for now...
    // TODO: think of a way to use it? or outlaw named fat-arrow expressions?
    var _fn = _types2.default.arrowFunctionExpression(params, body, async);
    if (node.returnType) _fn.returnType = node.returnType;
    if (node.typeParameters) _fn.typeParameters = node.typeParameters;
    return _fn;
  }
}

function replaceWithArrowFunction(path) {
  return path.replaceWith(toArrowFunction(path.node));
} // Replace a functional node with a new node binding the function to `this`.


function replaceWithBoundFunction(path) {
  if ((0, _is2.default)("Statement", path.node)) {
    // function f() {...}
    // f = f.bind(this)
    replaceWithPlainFunction(path);

    var bound = _types2.default.callExpression(_types2.default.memberExpression(path.node.id, _types2.default.identifier("bind")), [_types2.default.thisExpression()]);
    var assignToBound = _types2.default.expressionStatement(_types2.default.assignmentExpression("=", path.node.id, bound));

    return path.insertAfter(assignToBound);
  } else {
    // (function f() {...}).bind(this)
    var unbound = toPlainFunction(path.node);
    var _bound = _types2.default.callExpression(_types2.default.memberExpression(unbound, _types2.default.identifier("bind")), [_types2.default.thisExpression()]);
    return path.replaceWith(_bound);
  }
}

function isNamedArrowFunction(node) {
  // XXX: this seems a bit hackish.
  return typeof node.skinny === "boolean";
}